# Copyright (c) 2018-2021  University of Houston Networking Lab

from __future__ import absolute_import
import functools

class UnknownEdgeVertexError(Exception): # pylint: disable=C0111
  def __init__ (self, vertex):
    super(UnknownEdgeVertexError, self).__init__()
    self.v = vertex
  def __str__ (self): # pragma: no cover
    return "Vertex %s is not connected to this edge" % (str(self.v))

class GraphAlreadyAttachedError(Exception): # pylint: disable=C0111
  def __str__ (self): # pragma: no cover
    return "This object is already a member of a graph"

class UnknownConnectedVertexError(Exception): # pylint: disable=C0111
  def __init__ (self, sv, tv):
    super(UnknownConnectedVertexError, self).__init__()
    self.sv = sv
    self.tv = tv
  def __str__ (self): # pragma: no cover
    return "Vertex %s has no connection to target %s" % (str(self.sv), str(self.tv))


class GraphBase(object):
  def __init__ (self):
    self._edges = {}
    self._vertices = {} # { <id> : <vertex-obj>, ... }

  def addEdge (self, edge):
    edge._graphAttach(self)
    self._edges[edge._id] = edge
    return edge

  def removeEdges (self, edges):
    for edge in edges:
      edge._graph = None
      edge.disconnect()
      del self._edges[edge._id]

  def addVertex (self, vertex):
    vertex._graphAttach(self)
    self._vertices[vertex._id] = vertex
    return vertex

  def removeVertex (self, vertex):
    vertex._graph = None
    del self._vertices[vertex._id]

  def getDotString (self, as_subgraph = False):
    dot_data = []
    dda = dot_data.append
    if as_subgraph:
      dda("subgraph {")
    else:
      dda("graph {")

    for edge in self._edges.values():
      if as_subgraph:
        dda(""""%s.%d" -- "%s.%d" """ % (as_subgraph, edge._vertex_a._id, as_subgraph, edge._vertex_b._id))
      else:
        dda(""""%d" -- "%d" """ % (edge._vertex_a._id, edge._vertex_b._id))

    for vertex in self._vertices.values():
      if as_subgraph:
        dda(""""%s.%d" [label="%s",shape=%s,color=%s]""" % (as_subgraph, vertex._id, vertex.dotname, vertex.dotshape, vertex.dotcolor))
      else:
        dda(""""%d" [label="%s",shape=%s,color=%s]""" % (vertex._id, vertex.dotname, vertex.dotshape, vertex.dotcolor))

    dda("}")
    return "\n".join(dot_data)

  def getOrderedDotString (self):
    dot_data = []
    dda = dot_data.append

    for edge in self._edges.values():
      edg = [edge._vertex_a._id, edge._vertex_b._id]
      edg.sort()
      dda(""""%d" -- "%d" """ % (edg[0], edg[1]))

    for vertex in self._vertices.values():
      dda(""""%d" [label="%s"]""" % (vertex._id, vertex.dotname))

    dot_data.sort()
    dot_data.insert(0, "graph {")
    dda("}")
    return "\n".join(dot_data)


#@total_ordering
class VertexBase(object):
  ID = 0

  def __init__ (self):
    self._id = VertexBase.ID
    VertexBase.ID += 1
    self.dotshape = "oval"
    self.dotcolor = "black"

    self._graph = None
    self._edges = {}
    self._name = None
    self._dotname = None

  def __setitem__(self, key, value):
    self.__setattr__(key, value)

  def __setattr__(self, key, value):
    super().__setattr__(key, value)

  def __getitem__ (self, key):
    return self.__dict__[key]

  @property
  def name (self):
    if not self._name:
      return "_vertex-%d" % (self._id)
    return self._name

  @property
  def dotname (self):
    if not self._dotname:
      return self.name
    return self._dotname

  def __str__ (self):
    return "<%d>" % (self._id)

  def __eq__ (self, other):
#    return id(self) == id(other)
    return self._id == other._id

  def __lt__ (self, other):
    return self._id < other._id

  def _graphAttach (self, graph):
    if self._graph:
      raise GraphAlreadyAttachedError()
    self._graph = graph

  def addEdge (self, edge):
    self._edges[edge._id] = edge

  def removeEdge (self, edge):
    del self._edges[edge._id]

  def getEdge (self, other_vertex):
    for edge in self._edges.values():
      if edge.traverse(self) == other_vertex:
        return edge
    raise UnknownConnectedVertexError(self, other_vertex)

  def copyUnconnected (self):
    newv = self.__class__()
    newv._name = self._name
    newv._id = self._id
    return newv


class EdgeBase(object):
  ID = 0

  def __init__ (self):
    self._id = EdgeBase.ID
    EdgeBase.ID += 1
    self._graph = None
    self._vertex_a = None
    self._vertex_b = None

  def _graphAttach (self, graph):
    if self._graph:
      raise GraphAlreadyAttachedError()
    self._graph = graph
    self._vertex_a.addEdge(self)
    self._vertex_b.addEdge(self)

  def traverse (self, vertex):
    if vertex == self._vertex_a: # pylint: disable=R1705
      return self._vertex_b
    elif vertex == self._vertex_b:
      return self._vertex_a
    raise UnknownEdgeVertexError(vertex)

  def disconnect (self):
    self._vertex_a.removeEdge(self)
    self._vertex_b.removeEdge(self)
