# Copyright (c) 2018-2019  University of Houston Networking Lab

from __future__ import absolute_import

from . import base


class IllegalVLANIDError(Exception): # pylint: disable=C0111
  def __init__ (self, vid):
    super(IllegalVLANIDError, self).__init__()
    self._vid = vid
  def __str__ (self):
    return "%s is not a valid VLAN ID" % (str(self._vid))

class UnknownVertexNameError(Exception): # pylint: disable=C0111
  def __init__ (self, name):
    super(UnknownVertexNameError, self).__init__()
    self.name = name
  def __str__ (self):
    return "Vertex name (%s) is not found in this graph" % (self.name)

class OrphanPortError(Exception): # pylint: disable=C0111
  def __str__ (self):
    return "Port is not connected to a node"


class VLANSet(object):
  """Container for efficiently working with sets of vlan IDs"""

  class IDInIncludedRangeError(Exception): # pylint: disable=C0111
    def __init__ (self, vid):
      super(VLANSet.IDInIncludedRangeError, self).__init__()
      self._vid = vid
    def __str__ (self):
      return "VLAN ID %d is already included as part of a range" % (self._vid)

  class NegativeRangeError(Exception): # pylint: disable=C0111
    def __init__ (self, minvid, maxvid):
      super(VLANSet.NegativeRangeError, self).__init__()
      self._minvid = minvid
      self._maxvid = maxvid
    def __str__ (self):
      return "Minimum VLAN ID (%d) is larger than maximum VLAN ID (%d)" % (self._minvid, self._maxvid)

  def __init__ (self, ranges = None, ids = None):
    self._ranges = [] # [(min, max), ...]
    self._ids = set() # [ int, ...]

    if isinstance(ranges, (list, tuple)):
      for (vmin, vmax) in ranges:
        self.addRange(vmin, vmax)

    if isinstance(ids, (list, tuple)):
      for val in ids:
        self.addID(val)

  def addRange (self, vid_min, vid_max):
    """Add a VLAN ID range to the container

    Args:
      vlan_min (int): Low end of added range (inclusive)
      vlan_max (int): High end of added range (inclusive)

    Raises:
      IllegalVLANIDError: If either VLAN ID is less than 0 or greater than 4092
    """
    if vid_min > vid_max:
      raise VLANSet.NegativeRangeError(vid_min, vid_max)
    if vid_min < 0:
      raise IllegalVLANIDError(vid_min)
    if vid_max > 4092:
      raise IllegalVLANIDError(vid_max)
    self._ranges.append((vid_min, vid_max))

  def addID (self, vid):
    """Add a single VLAN ID to the container

    Args:
      vid (int): VLAN ID to add

    Raises:
      IllegalVLANIDError: If the VLAN ID is less than 0 or greater than 4092
      VLANSet.IDInIncludedRangeError: If the VLAN ID is already part of an included range
    """
    if vid < 0 or vid > 4092:
      raise IllegalVLANIDError(vid)
    for vmin,vmax in self._ranges:
      if vmin <= vid <= vmax:
        raise VLANSet.IDInIncludedRangeError(vid)
    self._ids.add(vid)


class L2Graph(base.GraphBase):
  """Basic layer-2 graph structure with some algorithms"""

  def __init__ (self):
    super(L2Graph, self).__init__()
    self._nodes_byname = {}

  def getDotString (self, include_ports = True): # pylint: disable=W0221
    if include_ports:
      return super(L2Graph, self).getDotString()

    dot_data = []
    dda = dot_data.append
    dda("graph {")

    visited_pairs = set()

    for vertex in self._vertices.values():
      if isinstance(vertex, Port):
        continue

      for port in vertex.portIter():
        rnode = port.getLinkedPort().getNode()
        idpair = [port._id, port.getLinkedPort()._id]
        idpair.sort()
        if tuple(idpair) in visited_pairs:
          continue

        dda(""""%d" -- "%d" """ % (vertex._id, rnode._id))
        visited_pairs.add(tuple(idpair))

      dda(""""%d" [label="%s",shape=%s]""" % (vertex._id, vertex.name, vertex.dotshape))

    dda("}")
    return "\n".join(dot_data)

  def addVertex (self, vertex):
    super(L2Graph, self).addVertex(vertex)
    if isinstance(vertex, Node):
      self._nodes_byname[vertex.name] = vertex
    return vertex

  def __getitem__ (self, key):
    try:
      return self._nodes_byname[key]
    except KeyError:
      raise UnknownVertexNameError(key)

  def prune (self):
    for node in self._nodes_byname.values():
      node.prune()


class Node(base.VertexBase):
  """Basic layer-2 node with rooted spanning tree support"""

  def __init__ (self, name = None):
    super(Node, self).__init__()
    self._name = name

  def __str__ (self):
    return "<%d: %s>" % (self._id, self._name)

  def portIter (self, filterfunc = None):
    for edge in self._edges.values():
      if filterfunc:
        if filterfunc(edge.port):
          yield edge.port
      else:
        yield edge.port

  def prune (self):
    """Remove ports that are unconnected to remote vertices"""

    for port in self.portIter():
      if port.unconnected:
        for edge_id in port._edges:
          del self._edges[edge_id]
        self._graph.removeEdges(port._edges.values())
        self._graph.removeVertex(port)

  def addPort (self, port):
    ie = InternalEdge(self, port)
    self._graph.addVertex(port)
    self._graph.addEdge(ie)
    return port

  def copyUnconnected (self, new_graph): # pylint: disable=W0221
    """Copy this node and its ports into a new graph without connected remote vertices"""

    newv = super(Node, self).copyUnconnected()
    new_graph.addVertex(newv)

    for port in self.portIter():
      newp = port.copyUnconnected()
      newv.addPort(newp) # pylint: disable=E1101

    return newv

  def generateSpanningTree (self):
    """Generate a new graph that is a spanning tree with this node as the root"""

    visited_vertices = set() # [vertex._id, ...]
    candidate_edges = set() # [[min(port._id), max(port._id)], ...]
    candidate_nodes = set() # [vertex._id, ...]
    next_vertices = [self]

    st = L2Graph()

    while next_vertices:
      cv = next_vertices.pop()
      visited_vertices.add(cv._id)

      if cv._id not in st._vertices:
        cv.copyUnconnected(st)

      for local_port in cv.portIter():
        remote_port = local_port.getLinkedPort()
        idpair = [local_port._id, remote_port._id]
        idpair.sort()

        if tuple(idpair) in candidate_edges:
          st_local_port = st._vertices[local_port._id]
          st_remote_port = st._vertices[remote_port._id]
          edge = LinkEdge(st_local_port, st_remote_port)
          st.addEdge(edge)
          continue

        remote_node_id = remote_port.getNode()._id
        if (remote_node_id in visited_vertices) or (remote_node_id in candidate_nodes):
          continue

        candidate_edges.add(tuple(idpair))
        candidate_nodes.add(remote_node_id)
        next_vertices.append(remote_port.getNode())

    st.prune()
    return st


class Port(base.VertexBase):
  """Base layer-2 port for layer-2 nodes"""

  def __init__ (self, name = None):
    super(Port, self).__init__()
    self.dotshape = "box"
    self._name = name
    self.client_id = None

  @property
  def unconnected (self):
    if len(self._edges) == 1:
      return True
    return False

  def copyUnconnected (self):
    newv = self.__class__(self._name)
    newv._id = self._id
    newv.client_id = self.client_id
    return newv

  def getLinkedPort (self):
    """Return the port connected to the other side of this one on a non-internal edge"""
    for edge in self._edges.values():
      if isinstance(edge, InternalEdge):
        continue
      return edge.traverse(self)

  def getNode (self):
    for edge in self._edges.values():
      if isinstance(edge, InternalEdge):
        return edge.node
    raise OrphanPortError()


class EthernetPort(Port):
  """Port specialization for Ethernet w/48-bit MAC Addresses"""

  def __init__ (self, name, macaddr):
    super(EthernetPort, self).__init__(name)
    self._macaddr = macaddr

  def copyUnconnected (self):
    newv = self.__class__(self._name, self._macaddr)
    newv._id = self._id
    newv.client_id = self.client_id
    return newv


class VLANEdge(base.EdgeBase):
  def __init__ (self, native_vid, allowed_vids):
    super(VLANEdge, self).__init__()
    self._native_vid = native_vid
    self._allowed_vids = allowed_vids


class InternalEdge(base.EdgeBase):
  """Edge subclass for edges between ports and nodes"""

  def __init__ (self, node, port):
    super(InternalEdge, self).__init__()
    self._vertex_a = node
    self._vertex_b = port

    self._vertex_a.addEdge(self)
    self._vertex_b.addEdge(self)

  @property
  def port (self):
    return self._vertex_b

  @property
  def node (self):
    return self._vertex_a


class LinkEdge(base.EdgeBase):
  """Edge subclass for edges between ports"""

  def __init__ (self, port_a, port_b):
    super(LinkEdge, self).__init__()
    self._vertex_a = port_a
    self._vertex_b = port_b

    self._vertex_a.addEdge(self)
    self._vertex_b.addEdge(self)
