# Copyright (c) 2019-2021  University of Houston Networking Lab

from __future__ import absolute_import
import copy
import importlib
import json
import math
import sys

from . import base
from ..log import logger


class MaxVerticesExceededError(Exception): # pylint: disable=C0111
  def __str__ (self):
    return "Attempted to connect more than two vertices to edge."

class CapacityExceededError(Exception): # pylint: disable=C0111
  def __init__ (self, vertex):
    super(CapacityExceededError, self).__init__()
    self.vertex = vertex
  def __str__ (self):
    return "Vertex %s capacity exceeded." % (str(self.vertex))

class CapacityLimitExceededError(Exception): # pylint: disable=C0111
  def __init__ (self, base_vertex, other_vertex):
    super(CapacityLimitExceededError, self).__init__()
    self.base_vertex = base_vertex
    self.other_vertex = other_vertex
  def __str__ (self):
    return "Vertex %s capacity limit exceeded." % self.base_vertex._name

class VertexAlreadyLoadedError(Exception): # pylint: disable=C0111
  def __init__ (self, base_vertex, loaded_vertex):
    super(VertexAlreadyLoadedError, self).__init__()
    self.base_vertex = base_vertex
    self.loaded_vertex = loaded_vertex
  def __str__ (self):
    return "Vertex %s already loaded onto vertex %s." % (self.loaded_vertex._name, self.base_vertex._name)

class VertexNotFoundNameError(Exception): # pylint: disable=C0111
  def __init__ (self, name):
    super(VertexNotFoundNameError, self).__init__()
    self.name = name
  def __str__ (self):
    return "Vertex not found by name: %s." % (self.name)

class UnknownVertexError(Exception): # pylint: disable=C0111
  def __init__ (self, vertex, unknown_vertex):
    super(UnknownVertexError, self).__init__()
    self.vertex = vertex
    self.unknown_vertex = unknown_vertex
  def __str__ (self):
    return "Vertex %s (id: %d) unknown on vertex %s (id: %d)" % (self.unknown_vertex._name, self.unknown_vertex._id, self.vertex._name, self.vertex._id)

class NonexistentVertexEdgeError(Exception): # pylint: disable=C0111
  def __init__ (self, edge):
    super(NonexistentVertexEdgeError, self).__init__()
    self.edge = edge
  def __str__ (self):
    return "Attempting to connect edge (id: %d) without vertices." % self.edge._id


class GraphBase(base.GraphBase):
  def __init__ (self, name = None):
    super(GraphBase, self).__init__()
    self.name = name
    self._edgecache = set()
    self._vtypecache = {}
    self._lcvtype = None

  @property
  def vtypecache (self):
    if not self._vtypecache:
      for vertex in self._vertices.values():
        self._vtypecache.setdefault(type(vertex), []).append(vertex)
    return self._vtypecache

  def getVertexListByType (self, typ):
    try:
      return self.vtypecache[typ]
    except KeyError:
      return []

  def getVertexByName (self, name):
    for vertex in self._vertices.values():
      if vertex._name == name:
        return vertex
    raise VertexNotFoundNameError(name)

  def neighbors (self, node):
    for vertex in self._vertices[node].getNeighbors():
      yield vertex._id

  @property
  def nodes (self):
    return self._vertices

  @property
  def edges (self):
    for edge in self._edges.values():
      yield (edge._vertex_a._id, edge._vertex_b._id)

  @property
  def edgecache (self):
    if not self._edgecache:
      for edge in self._edges.values():
        eh = None
        if edge._vertex_a._id < edge._vertex_b._id:
          eh = (edge._vertex_a._id, edge._vertex_b._id)
        else:
          eh = (edge._vertex_b._id, edge._vertex_a._id)
        self._edgecache.add(eh)
    return self._edgecache

  def edgeExists (self, va, vb):
    edge = None
    if va < vb:
      edge = (va,vb)
    else:
      edge = (vb,va)

    return edge in self.edgecache

  def lowestCountVertexType (self):
    """Returns the type of the lowest count vertex. Ties broken arbitrarily."""    
    if not self._lcvtype:
      minimum = math.inf
      for typ, lst in self.vtypecache.items():
        count = len(lst)
        if count < minimum:
          minimum = count
          self._lcvtype = typ
    return self._lcvtype

  def getEdgeEndpointTypes (self):
    edge_types = set()
    for edge in self._edges.values():
      edge_types.add(edge.vertexTypes())
    return edge_types

  def __json__ (self):
    g = {}
    g["vertices"] = [vertex.__json__() for vertex in self._vertices.values()]
    g["edges"] = [edge.__json__() for edge in self._edges.values()]

    if self.name:
      g["name"] = self.name

    return g

  def __dimacs__ (self, types = False):
    dimacs_list = []
    vertex_dict = {}
    for idx, vertex in enumerate(self._vertices.values()):
      vertex_dict[vertex._id] = (idx + 1, vertex)

    dimacs_list.append("%d %d" % (len(self._vertices), len(self._edges)))

    for vertex_id, tup in vertex_dict.items():
      num = tup[0]
      v = tup[1]

      if types:
#        line = "%d %s" % (v._capacity, str(type(v)))
        line = "%s %d" % (v.type_name, v._capacity)
      else:
        line = "%d" % v._capacity

      for neighbor in tup[1].getNeighbors():
        line += " %d" % vertex_dict[neighbor._id][0]
      dimacs_list.append(line)
    
    return "\n".join(dimacs_list)

  def __netx__ (self):
    import networkx as nx

    g = nx.Graph()
    node_map = {}
    node_list = sorted(list(self._vertices.keys()))

    for idx, node in enumerate(node_list):
      node_map[node] = idx
      g.add_node(idx)
      g.nodes[idx]["type"] = self._vertices[node].type_name

    for edge in self._edges.values():
      g.add_edge(node_map[edge._vertex_a._id], node_map[edge._vertex_b._id])
      
    #for node in g.nodes:
    #  cap = 10
    #  ig.nodes[node]["capacity"] = cap
    #  ig.nodes[node]["remaining_cap"] = cap

    return g


class InfrastructureGraph(GraphBase):
  def __init__ (self, name = None):
    super(InfrastructureGraph, self).__init__(name)
    logger.debug("NETUTILS IG LOG MESSAGE")
      
  def addLoad (self, solution_graph, request_graph):
    for ig_node_id, rg_node_ids in solution_graph._vertices.items():
      for rg_node_id in rg_node_ids:
        rg_node = request_graph._vertices[rg_node_id]
        self._vertices[ig_node_id].addLoad(rg_node)

  def delLoad (self, solution_graph, request_graph):
    for ig_node_id, rg_node_ids in solution_graph._vertices.items():
      for rg_node_id in rg_node_ids:
        rg_node = request_graph._vertices[rg_node_id]
        self._vertices[ig_node_id].delLoad(rg_node)

  def printStatus (self):
    print("Current graph load...")
    for vertex in self._vertices.values():
      print(vertex._name, "%.4f" % vertex.getLoad())
    print()

  def getLoad (self):
    util = []
    names = []
    for node in self._vertices.values():
      util.append(node.getLoad())
      names.append(node._name)
    return names, util

#  def __netx__ (self):
#    g = super.__netx__(self)


class RequestGraph(GraphBase):
  def __init__ (self, name = None):
    super(RequestGraph, self).__init__(name)

  def renderSolutionDotString (self, matches, as_subgraph = False):
    dot_data = []
    dda = dot_data.append
    if as_subgraph:
      dda("subgraph {")
    else:
      dda("graph {")

    for edge in self._edges.values():
      if as_subgraph:
        dda(""""%s.%d" -- "%s.%d" """ % (as_subgraph, edge._vertex_a._id, as_subgraph, edge._vertex_b._id))
      else:
        dda(""""%d" -- "%d" """ % (edge._vertex_a._id, edge._vertex_b._id))

    for vertex in self._vertices.values():
      name = "ig-%d" % (matches[vertex._id])

      if as_subgraph:
        dda(""""%s.%d" [label="%s",shape=%s,color=%s]""" % (as_subgraph, vertex._id, name, vertex.dotshape, vertex.dotcolor))
      else:
        dda(""""%d" [label="%s",shape=%s,color=%s]""" % (vertex._id, name, vertex.dotshape, vertex.dotcolor))

    dda("}")
    return "\n".join(dot_data)

class SolutionGraph(object):
  def __init__ (self):
    self._vertices = {} # {<int> : [<int>, ...], ...}

  def getPackingNode (self, other):
    for infra_node, req_nodes in self._vertices.items():
      if other._id in req_nodes:
        return infra_node

  def addNewPackingVertex (self, rg_neighbor, ig_neighbor):
    try:
      v = self._vertices[ig_neighbor._id]
      v.append(rg_neighbor._id)
#      self._vertices[ig_neighbor._id].append(rg_neighbor._id)
    except KeyError:
      self._vertices[ig_neighbor._id] = [rg_neighbor._id]

  def copy (self):
    s = SolutionGraph()
    s._vertices = {k:v[:] for k,v in self._vertices.items()}
    return s

  def __str__ (self):
    out_str = []
    ig_nodes = list(self._vertices.keys())
    out_str.append(str(len(ig_nodes)))
    ig_nodes.sort()
    for node in ig_nodes:
      for rg_node in sorted(self._vertices[node]):
        out_str.append("%d,%d" % (node, rg_node))
    return "\n".join(out_str)

  @property
  def lower_bound (self):
    return len(self._vertices.keys())

  @property
  def match (self):
    matches = {} # {rg_node : ig_node ... }
    for ig_node, rg_nodes in self._vertices.items():
      for rg_node in rg_nodes:
        matches[rg_node] = ig_node
    return matches

class VertexBase(base.VertexBase):
  def __init__ (self, cap, name = None):
    super(VertexBase, self).__init__()
    self._total_capacity = cap
    self._usable_capacity = cap
    self._capacity = cap
    self._name = name
    self._loaded = []
    self.type_name = "base"
    self._neighbors = []

#  def __getattr__ (self, key):
#    if not self.__dict__.has_key(key):
#      value = self.fetchAttr(key)
#      self.__dict__[key] = value
#    return self.__dict__[key]
#
#  def __setattr__ (self, key, value):
#    if key not in self.__dict__.keys():
#      #value = self.fetchAttr(key)
#      self.__dict__[key] = value

  def __lt__ (self, other):
    return self.name < other.name

  def __hash__ (self):
    return self._id

  @property
  def _dotname (self):
    return self.name

  @_dotname.setter
  def _dotname (self, val):
    pass

  @property
  def used_capacity (self):
    return self._total_capacity - self._capacity

  @property
  def neighbors (self):
    if not self._neighbors:
      self._neighbors = [edge.traverse(self) for edge in self._edges.values()]
    return self._neighbors

  def addEdge (self, edge):
    super(VertexBase, self).addEdge(edge)
    self._neighbors = []

  def available_capacity (self):
    return self._usable_capacity - self.used_capacity
    #return self._capacity - (self._total_capacity - self._usable_capacity)

  def applyCapacityLimit (self, limit): # as a percent
    self._usable_capacity = self._total_capacity * limit

  def addLoad (self, vertex):
    if vertex in self._loaded:
      raise VertexAlreadyLoadedError(self, vertex)
    if vertex._capacity > self.available_capacity():
      raise CapacityLimitExceededError(self, vertex)
    logger.debug("addLoad -- Node %s, other %s" % (self._name, vertex._name))

    self._capacity -= vertex._capacity
    self._loaded.append(vertex)
    self._packing_vertices = []

  def delLoad (self, vertex):
    if vertex not in self._loaded:
      raise UnknownVertexError(self, vertex)
    if (self._capacity + vertex._capacity) > self._total_capacity:
      raise CapacityExceededError(self)
    logger.debug("delLoad -- Node %s, other %s" % (self._name, vertex._name))

    self._capacity += vertex._capacity
    self._loaded.remove(vertex)

  def getLoad (self):
    return (self.used_capacity / self._total_capacity)

  def getNeighbors(self):
    return self.neighbors

  def getEdgelessCopy (self):
    obj_type = type(self)
    obj = obj_type(self._total_capacity, name = self._name)
    obj._id = self._id

    return obj

  def __json__ (self):
    return { "id" : self._id, 
             "name" : self._name, 
             "cap" : self._capacity,
             "type" : self.type_name,
             "class" : self.__class__.__name__,
             "module" : self.__module__}


class SolutionVertex(VertexBase):
  def __init__ (self, ig_node, rg_node):
    super(SolutionVertex, self).__init__(0, "%s-%s" % (ig_node._name, rg_node._name))
    self.ig_node = ig_node
    self.rg_node = rg_node
 

class RedVertex(VertexBase):
  def __init__ (self, cap, name = None):
    super(RedVertex, self).__init__(cap, name)
    self.dotcolor = "red"
    self.type_name = "red"
    

class BlueVertex(VertexBase):
  def __init__ (self, cap, name = None):
    super(BlueVertex, self).__init__(cap, name)
    self.dotcolor = "blue"
    self.type_name = "blue"
 

class GreenVertex(VertexBase):
  def __init__ (self, cap, name = None):
    super(GreenVertex, self).__init__(cap, name)
    self.dotcolor = "green"
    self.type_name = "green"


class PurpleVertex(VertexBase):
  def __init__ (self, cap, name = None):
    super(PurpleVertex, self).__init__(cap, name)
    self.dotcolor = "purple"
    self.type_name = "purple"


class CPUVertex(VertexBase):
  def __init__ (self, cap, name = None):
    super(CPUVertex, self).__init__(cap, name)
    self.dotcolor = "black"
    self.type_name = "cpu"


class RAMVertex(VertexBase):
  def __init__ (self, cap, name = None):
    super(RAMVertex, self).__init__(cap, name)
    self.dotcolor = "black"
    self.type_name = "ram"


class DiskVertex(VertexBase):
  def __init__ (self, cap, name = None):
    super(DiskVertex, self).__init__(cap, name)
    self.dotcolor = "black"
    self.type_name = "disk"


class PortVertex(VertexBase):
  def __init__ (self, cap, name = None):
    super(PortVertex, self).__init__(cap, name)
    self.dotcolor = "black"
    self.type_name = "port"


class DataPlaneVertex(VertexBase):
  def __init__ (self, cap, name = None):
    super(DataPlaneVertex, self).__init__(cap, name)
    self.dotcolor = "black"
    self.type_name = "dp"


class EdgeBase(base.EdgeBase):
  def __init__ (self, vertex = None, vertex2 = None):
    super(EdgeBase, self).__init__()
    self._vertex_a = vertex
    self._vertex_b = vertex2

  def _attachToVertex(self):
    if self._vertex_a:
      self._vertex_a.addEdge(self)
    if self._vertex_b:
      self._vertex_b.addEdge(self)
    if not self._vertex_a and not self._vertex_b:
       raise NonexistentVertexEdgeError(self)
    return

  def connect (self, vertex, vertex2):
    if self._vertex_a or self._vertex_b:
      raise MaxVerticesExceededError()
    self._vertex_a = vertex
    self._vertex_b = vertex2
    self._attachToVertex()
    return

  def addVertex (self, vertex):
    if not self._vertex_a:
      self._vertex_a = vertex
    elif not self._vertex_b:
      self._vertex_b = vertex
    else:
      raise MaxVerticesExceededError()
    self._attachToVertex()
    return

  def getNodelessCopy (self):
    obj_type = type(self)
    obj = obj_type()
    obj._id = self._id
    return obj

  def vertexTypes (self):
    return tuple(sorted([self._vertex_a.type_name, self._vertex_b.type_name]))

  def __json__ (self):
    d = { "id" : self._id } 
    if self._vertex_a:
      d["vertex_a"] = self._vertex_a._id
    if self._vertex_b: 
      d["vertex_b"] = self._vertex_b._id 
    return d


def vertexTypesSet(vertices):
  return set([type(v) for v in vertices])
#  types = set()
#  for vertex in vertices:
#    types.add(type(vertex))
#  return types
  

def graph_factory (cls, obj):
  try:
    name = obj["name"]
  except KeyError:
    name = None

  g = cls(name)

  for v in obj["vertices"]:
    mod = importlib.import_module(v["module"])
    vertex_type = getattr(mod, v["class"])
    vertex = vertex_type(v["cap"], name = v["name"])
    vertex._id = v["id"]
    g.addVertex(vertex)

  for e in obj["edges"]:
    edge = EdgeBase(g._vertices[e["vertex_a"]], g._vertices[e["vertex_b"]])
    edge._id = int(e["id"])
    g.addEdge(edge)
 
  return g
