# Copyright (c) 2019  University of Houston

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import random

from IPy import IP

### Utils for IP subnet randomization

def getRandomNetwork (prefix_len, exclude_nets = None):
  if exclude_nets is None:
    exclude_nets = []
  not_found = True
  while not_found:
    not_found = False
    nint = random.randint(1, 3758096383)
    for net in exclude_nets:
      if nint in net:
        not_found = True
        break
  ip1 = IP(nint)
  ip2 = IP("%s/%d" % (ip1.strNormal(), prefix_len), make_net = True)
  return ip2

def getSubnets (min_prefix, max_prefix, count, exclude_nets = None):
  subnets = []
  while len(subnets) != count:
    plen = random.randint(min_prefix, max_prefix)
    net = getRandomNetwork(plen, exclude_nets)
    try:
      for subnet in subnets:
        if net.overlaps(subnet):
          raise ValueError
    except ValueError:
      continue
    subnets.append(net)
  return subnets

class InsufficientHostSpaceError(Exception):
  """Exception raised for errors when there is not enough number
  of host IP addresses in the subnet address range given.
  """
  def __init__ (self,num_hosts, net_len):
    super(InsufficientHostSpaceError, self).__init__()
    self._num_hosts = num_hosts
    self._net_len = net_len

  def __str__ (self):
    return "There are not sufficient number of IPs (= %d) for this many host IPs (= %d)" % (self._net_len, self._num_hosts)


def getNetworkHosts (net, num_hosts, exclude_ips = None):
  hostIPs = set([])
  max_num_ips = net.len() - 2
  if num_hosts > max_num_ips:
    raise InsufficientHostSpaceError(num_hosts, max_num_ips)

  if exclude_ips is None:
    exclude_ips = set()

  for i in range(num_hosts):
    # reserve exclude set
    new_ip = net[random.randint(1, net.len()-2)]
    while (new_ip in hostIPs) or (new_ip in exclude_ips):
      new_ip = net[random.randint(1, net.len()-2)]
    hostIPs.add(new_ip)

  return list(hostIPs)

def randomNames (name_prefix, no_names = 2, rand_range = 20):
  NAMEList = []
  for idx in range(1, rand_range):
    NAMEList.append("%s%02d" % (name_prefix, idx))
  random.shuffle(NAMEList)
  return NAMEList[0:no_names]

class NetworkSequence(object):
  def __init__ (self, subnetstr, prefix_len):
    self.subnet = IP(subnetstr, make_net=True)
    self.prefix_len = prefix_len

  def __iter__ (self):
    return NetSeqIter(self)


class NetSeqIter(object):
  def __init__ (self, ns):
    self._ns = ns
    self.iter = 0
    self.netsize = 2**(32-self._ns.prefix_len)

  def __iter__ (self):
    return self

  def next (self):
    try:
      ret = IP("%s/%d" % (self._ns.subnet[self.iter * self.netsize], self._ns.prefix_len))
      self.iter += 1
      return ret
    except IndexError:
      raise StopIteration

  def __next__ (self):
    return self.next()
