# Copyright (c) 2015-2016  Barnstormer Softworks, Ltd.
# Copyright (c) 2017-2021  University of Houston Networking Lab

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import binascii
import socket
import struct

import six

if six.PY3: # Xpy2
  from builtins import int # pylint: disable=E0401
  BORD = int
else: # Xpy3
  BORD = ord

class DPID(object):
  """Utility class representing OpenFlow Datapath IDs

  This class tries to handle all likely inputs and desired outputs, while
  providing a single internal type to work with in the code.

  String representations passed in must be represented in hex, but may contain
  common separators (colon, dash, and period) in any configuration.

  Args:
    val (int, unicode, str)

  Raises:
    DPID.OutOfRangeError: If the DPID represented by `val` is larger than the spec allows
      or less than zero.
    DPID.InputTypeError: If `val` is not a supported data type
  """

  MAX = (2 ** 64) - 1

  class OutOfRangeError(Exception): # pylint: disable=C0111
    def __init__ (self, val):
      super(DPID.OutOfRangeError, self).__init__()
      self.val = val
    def __str__ (self): # pragma: no cover
      return "Input value (%d) out of range of valid DPIDs" % (self.val)

  class InputTypeError(Exception): # pylint: disable=C0111
    def __init__ (self, val):
      super(DPID.InputTypeError, self).__init__()
      self.val = val
    def __str__ (self): # pragma: no cover
      return "Input value (%s) has invalid type (%s)" % (self.val, type(self.val))

  def __init__ (self, val):
    self._dpid = None

    if isinstance(val, (six.string_types)):
      val = val.replace(".", "")
      val = val.replace(":", "")
      val = val.replace("-", "")
      val = int(val, 16)

    if isinstance(val, (six.integer_types)):
      if 0 <= val < DPID.MAX:
        self._dpid = val
      else:
        raise DPID.OutOfRangeError(val)
    else:
      raise DPID.InputTypeError(val)

  def __eq__ (self, other):
    return self._dpid == other._dpid # pylint: disable=W0212

  def __hash__ (self):
    return self._dpid

  def __str__ (self):
    """
    Returns:
      str: Hex formatted DPID, with colons
    """
    s = self.hexstr()
    return ":".join(["%s%s" % (s[x], s[x+1]) for x in range(0,15,2)])

  def __repr__ (self):
    return str(self)

  def __json__ (self):
    return str(self)

  def hexstr (self):
    """Unformatted hex representation of DPID

    Returns:
      str: Hex formatted DPID, without colons
    """
    return "%016x" % (self._dpid)

class EthernetMAC (object):
  """Utility class representing 48-bit Ethernet MAC Addresses

  This class tries to handle all likely inputs and desired outputs, while
  providing a single internal type to work with in the code.

  String representations passed in must be represented in hex, but may contain
  common separators (colon, dash, and period) in any configuration.

  Args:
    val (int, unicode, str)

  Raises:
    EthernetMAC.OutOfRangeError: If the MAC represented by `val` is larger than
      than 48-bits or less than zero.
    EthernetMAC.InputTypeError: If `val` is not a supported data type
  """

  MAX = 2 ** 48

  class OutOfRangeError(Exception): # pylint: disable=C0111
    def __init__ (self, val):
      super(EthernetMAC.OutOfRangeError, self).__init__()
      self.val = val
    def __str__ (self): # pragma: no cover
      return "Input value (%d) out of range of valid Ethernet Addresses" % (self.val)

  class InputTypeError(Exception): # pylint: disable=C0111
    def __init__ (self, val):
      super(EthernetMAC.InputTypeError, self).__init__()
      self.val = val
    def __str__ (self): # pragma: no cover
      return "Input value (%s) has invalid type (%s)" % (self.val, type(self.val))

  def __init__ (self, val):
    self._mac = None

    if isinstance(val, (six.string_types)):
      if len(val) == 6: # Xpy3
        val = binascii.hexlify(val)
      val = val.replace(".", "")
      val = val.replace(":", "")
      val = val.replace("-", "")
      val = int(val, 16)
    elif isinstance(val, bytes): # Xpy2
      res = 0
      for byte in val:
        res = res * 256 + int(byte)
      val = res

    if isinstance(val, (int)):
      if 0 <= val < EthernetMAC.MAX:
        self._mac = val
      else:
        raise EthernetMAC.OutOfRangeError(val)
    else:
      raise EthernetMAC.InputTypeError(val)

  def __eq__ (self, other):
    return self._mac == other._mac # pylint: disable=W0212

  def __hash__ (self):
    return self._mac

  def __str__ (self):
    """
    Returns:
      str: Hex formatted MAC, with colons
    """
    s = self.hexstr()
    return ":".join(["%s%s" % (s[x], s[x+1]) for x in range(0,11,2)])

  def __json__ (self):
    return str(self)

  def __repr__ (self):
    return str(self)

  def hexstr (self):
    """Unformatted hex representation of MAC

    Returns:
      str: Hex formatted MAC, without separators
    """
    return "%012x" % (self._mac)


class IPv4Address (object):
  """Utility class representing IPv4 Addresses

  This class tries to handle all likely inputs and desired outputs, while
  providing a single internal type to work with in the code.

  String representations must be passed in as dotted quads.

  Args:
    val (int, unicode, str)

  Raises:
    IPv4Address.OutOfRangeError: If the address represented by `val` is larger than
      than 32-bits or less than zero.
    IPv4Address.InputTypeError: If `val` is not a supported data type
  """

  MAX = (2 ** 32) - 1

  class OutOfRangeError(Exception): # pylint: disable=C0111
    def __init__ (self, val):
      super(IPv4Address.OutOfRangeError, self).__init__()
      self.val = val
    def __str__ (self): # pragma: no cover
      return "Input value (%d) out of range of valid IPv4 Addresses" % (self.val)

  class InputTypeError(Exception): # pylint: disable=C0111
    def __init__ (self, val):
      super(IPv4Address.InputTypeError, self).__init__()
      self.val = val
    def __str__ (self): # pragma: no cover
      return "Input value (%s) has invalid type (%s)" % (self.val, type(self.val))

  class InvalidFormatError(Exception): # pylint: disable=C0111
    def __init__ (self, val):
      super(IPv4Address.InvalidFormatError, self).__init__()
      self.val = val
    def __str__ (self): # pragma: no cover
      return "Format of input string (%s) is not a valid IPv4 address" % (self.val)

  def __init__ (self, val):
    self._ip = None

    if isinstance(val, (six.string_types)):
      if len(val) != 4:
        try:
          if val.count("."):
            self._ip = struct.unpack(">I", socket.inet_aton(val))[0]
        except (OSError, socket.error):
          raise IPv4Address.InvalidFormatError(val)

    if isinstance(val, six.binary_type) and not self._ip:
      res = 0
      for byte in val:
        res = res * 256 + BORD(byte)
      val = res

    if isinstance(val, (int)):
      if 0 <= val <= IPv4Address.MAX:
        self._ip = val
      else:
        raise IPv4Address.OutOfRangeError(val)

    if not self._ip:
      raise IPv4Address.InputTypeError(val)

  def __eq__ (self, other):
    return self._ip == other._ip # pylint: disable=W0212

  def __hash__ (self):
    return self._ip

  def __str__ (self):
    return socket.inet_ntoa(struct.pack("!I", self._ip))

  def __json__ (self):
    return str(self)

  def __repr__ (self):
    return str(self)
