# Copyright (c) 2019-2021 University of Houston Networking Lab

import pytest

import uhnet.graph.base

### Utils

def zero_ids ():
  uhnet.graph.base.VertexBase.ID = 0
  uhnet.graph.base.EdgeBase.ID = 0

### Fixtures

@pytest.fixture
def simple_graph ():
  zero_ids()

  g = uhnet.graph.base.GraphBase()

  v1 = uhnet.graph.base.VertexBase()
  v1._name = "v1"

  v2 = uhnet.graph.base.VertexBase()
  v2._name = "v2"

  v3 = uhnet.graph.base.VertexBase()

  e1 = uhnet.graph.base.EdgeBase()
  e1._vertex_a = v1
  e1._vertex_b = v2

  e2 = uhnet.graph.base.EdgeBase()
  e2._vertex_a = v1
  e2._vertex_b = v3

  g.addVertex(v1)
  g.addVertex(v2)
  g.addVertex(v3)
  g.addEdge(e1)
  g.addEdge(e2)

  return g


### Tests

def test_addremovevertex ():
  g = uhnet.graph.base.GraphBase()
  v1 = uhnet.graph.base.VertexBase()
  v2 = uhnet.graph.base.VertexBase()

  g.addVertex(v1)
  g.addVertex(v2)

  assert(len(g._vertices) == 2)

  g.removeVertex(v1)

  assert(len(g._vertices) == 1)
  assert(g._vertices[1])

  v3 = uhnet.graph.base.VertexBase()
  assert(v3._id == 2)


def test_dotstr (simple_graph):
  DOTLIST = """graph {\n"0" -- "1" \n"0" -- "2" \n"0" [label="v1",shape=oval,color=white]\n"1" [label="v2",shape=oval,color=white]\n"2" [label="_vertex-2",shape=oval,color=white]\n}""".split("\n")
  gdotl = simple_graph.getDotString().split("\n")
  print(gdotl)
  assert(len(DOTLIST) == len(gdotl))

  for (ref,graph) in zip(DOTLIST, gdotl):
    assert(ref.strip() == graph.strip())


def test_traverse1 (simple_graph):
  v1 = simple_graph._vertices[0]
  v2 = simple_graph._vertices[1]

  e1 = v2.getEdge(v1)
  assert(e1.traverse(v2) == v1)

def test_traverse2 (simple_graph):
  v1 = simple_graph._vertices[0]
  v2 = simple_graph._vertices[1]
  va = uhnet.graph.base.VertexBase()

  e1 = v2.getEdge(v1)
  with pytest.raises(uhnet.graph.base.UnknownEdgeVertexError):
    e1.traverse(va)

def test_vertexcopy (simple_graph):
  v2 = simple_graph._vertices[1]
  newv = v2.copyUnconnected()
  assert(len(newv._edges) == 0)
  assert(newv.name == v2.name)
  assert(newv._id == v2._id)

def test_vertexstr1 (simple_graph):
  v1 = simple_graph._vertices[0]
  v3 = simple_graph._vertices[2]

  assert(str(v1) == "<0>")
  assert(str(v3) == "<2>")


def test_vertexname1 (simple_graph):
  v2 = simple_graph._vertices[1]
  v3 = simple_graph._vertices[2]

  assert(v2.name == "v2")
  assert(v3.name == "_vertex-2")


def test_removeedge (simple_graph):
  v1 = simple_graph._vertices[0]
  v2 = simple_graph._vertices[1]

  e1 = v2.getEdge(v1)
  assert(len(v1._edges) == 2)

  simple_graph.removeEdges([e1])
  assert(len(v1._edges) == 1)


def test_unknownedge (simple_graph):
  v1 = simple_graph._vertices[0]
  va = uhnet.graph.base.VertexBase()

  with pytest.raises(uhnet.graph.base.UnknownConnectedVertexError):
    v1.getEdge(va)

def test_reattach1 (simple_graph):
  v1 = simple_graph._vertices[0]

  with pytest.raises(uhnet.graph.base.GraphAlreadyAttachedError):
    simple_graph.addVertex(v1)

def test_reattach2 (simple_graph):
  v1 = simple_graph._vertices[0]
  v2 = simple_graph._vertices[1]

  e1 = v2.getEdge(v1)
  with pytest.raises(uhnet.graph.base.GraphAlreadyAttachedError):
    simple_graph.addEdge(e1)
