# Copyright (c) 2019-2021 University of Houston Networking Lab

from __future__ import unicode_literals

import pytest

import uhnet.types

### DPID

def test_dpidmax ():
  with pytest.raises(uhnet.types.DPID.OutOfRangeError):
    uhnet.types.DPID(2**64)

def test_dpidmin ():
  with pytest.raises(uhnet.types.DPID.OutOfRangeError):
    uhnet.types.DPID(-1)

def test_dpidinvalid1 ():
  with pytest.raises(uhnet.types.DPID.InputTypeError):
    uhnet.types.DPID([0, 1, 2, 3])

def test_dpidstr1 ():
  dpid = uhnet.types.DPID("00-00.01:02-03:04:05:06")
  assert(str(dpid) == "00:00:01:02:03:04:05:06")

def test_dpideq1 ():
  dpid1 = uhnet.types.DPID(2)
  dpid2 = uhnet.types.DPID("00:00:00:00:00:00:00:02")
  assert(dpid1 == dpid2)

def test_dpidhash1 ():
  dpid1 = uhnet.types.DPID(2)
  dpid2 = uhnet.types.DPID("00:00:00:00:00:00:00:02")
  dpid3 = uhnet.types.DPID("00-00.01:02-03:04:05:06")
  assert(hash(dpid1) == hash(dpid2))
  assert(hash(dpid1) != hash(dpid3))

def test_dpidrepr1 ():
  dpid1 = uhnet.types.DPID(2)
  assert(repr(dpid1) == "00:00:00:00:00:00:00:02")

def test_dpidjson1 ():
  dpid1 = uhnet.types.DPID(2)
  assert(dpid1.__json__() == "00:00:00:00:00:00:00:02")

### EthernetMAC

def test_ethmax ():
  with pytest.raises(uhnet.types.EthernetMAC.OutOfRangeError):
    uhnet.types.EthernetMAC((2**48)+1)

def test_ethmin ():
  with pytest.raises(uhnet.types.EthernetMAC.OutOfRangeError):
    uhnet.types.EthernetMAC(-1)

def test_ethinvalid1 ():
  with pytest.raises(uhnet.types.EthernetMAC.InputTypeError):
    uhnet.types.EthernetMAC([1,2,3])

def test_ethstr1 ():
  mac = uhnet.types.EthernetMAC("ff-ff-ff-ff-ff-ff")
  assert(str(mac) == "ff:ff:ff:ff:ff:ff")

def test_ethstr2 ():
  mac = uhnet.types.EthernetMAC("ff:ff:ff:ff:ff:ff")
  assert(str(mac) == "ff:ff:ff:ff:ff:ff")

def test_ethstr3 ():
  mac = uhnet.types.EthernetMAC("ff.ff.ff.ff.ff.ff")
  assert(str(mac) == "ff:ff:ff:ff:ff:ff")

def test_ethstr4 ():
  mac = uhnet.types.EthernetMAC("FF-FF-FF-FF-FF-FF")
  assert(str(mac) == "ff:ff:ff:ff:ff:ff")

def test_ethstr5 ():
  mac = uhnet.types.EthernetMAC("aA-bb:Cc.dd-ee:f0")
  assert(str(mac) == "aa:bb:cc:dd:ee:f0")

def test_ethbytes1 ():
  mac = uhnet.types.EthernetMAC(b"\xff\xff\xff\xfe\xfe\xfe")
  assert(str(mac) == "ff:ff:ff:fe:fe:fe")

def test_hexstr1 ():
  mac = uhnet.types.EthernetMAC("aA-bb:Cc.dd-ee:f0")
  assert(mac.hexstr() == "aabbccddeef0")

def test_macrepr1 ():
  mac = uhnet.types.EthernetMAC("aA-bb:Cc.dd-ee:f0")
  assert(repr(mac) == "aa:bb:cc:dd:ee:f0")

def test_macjson1 ():
  mac = uhnet.types.EthernetMAC("aA-bb:Cc.dd-ee:f0")
  assert(mac.__json__() == "aa:bb:cc:dd:ee:f0")

def test_ethhash1 ():
  mac1 = uhnet.types.EthernetMAC("aA-bb:Cc.dd-ee:f0")
  mac2 = uhnet.types.EthernetMAC("aA-bb:Cc.dd-ee:f0")
  assert(hash(mac1) == hash(mac2))


### IPv4Address

def test_ipv4max ():
  with pytest.raises(uhnet.types.IPv4Address.OutOfRangeError):
    uhnet.types.IPv4Address(2**32)

def test_ipv4min ():
  with pytest.raises(uhnet.types.IPv4Address.OutOfRangeError):
    uhnet.types.IPv4Address(-1)

def test_ipv4_invalid1 ():
  with pytest.raises(uhnet.types.IPv4Address.InvalidFormatError):
    uhnet.types.IPv4Address("192.168.10.256")

def test_ipv4_invalid2 ():
  with pytest.raises(uhnet.types.IPv4Address.InvalidFormatError):
    uhnet.types.IPv4Address("192-168.10.255")

def test_ipv4_invalid3 ():
  with pytest.raises(uhnet.types.IPv4Address.InputTypeError):
    uhnet.types.IPv4Address([1,2,3,4])

def test_ipv4_int1 ():
  ip = uhnet.types.IPv4Address(1)
  assert(ip._ip == 1)
  assert(str(ip) == "0.0.0.1")

def test_ipv4_str1 ():
  ip = uhnet.types.IPv4Address("0.0.0.2")
  assert(ip._ip == 2)
  assert(str(ip) == "0.0.0.2")

def test_ipv4_repr1 ():
  ip = uhnet.types.IPv4Address("0.0.0.2")
  assert(ip._ip == 2)
  assert(repr(ip) == "0.0.0.2")

def test_ipv4_bytes1 ():
  ip = uhnet.types.IPv4Address(b"\x0e\x0d\x01\x01")
  assert(str(ip) == "14.13.1.1")

def test_ipv4_json1 ():
  ip = uhnet.types.IPv4Address("0.0.0.2")
  assert(ip._ip == 2)
  assert(ip.__json__() == "0.0.0.2")

def test_ipv4_hash1 ():
  ip1 = uhnet.types.IPv4Address("0.0.0.2")
  ip2 = uhnet.types.IPv4Address("0.0.0.2")

  assert(hash(ip1) == hash(ip2))


